% Copyright  2012 Computational Biomedicine Lab (CBL), 
% University of Houston. All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, is prohibited without the prior written consent of CBL.
%

function data = genExperiment(data)
% This function generates a struct needed to run the experiments necessary
% for parameter optimization.
% 
% INPUT:
%  - data: Oputput struct from the genIdx function.
%  - data: Same input as for genIdx.
%  - data: Additional options that are needed also for genFeats.
%     * featType: List of types of features that will be generated.
%        + 'LP':    Low Pass.
%        + 'HP':    High Pass.
%        + 'Lap':   Laplacian.
%        + 'Shear': Shearlets.
%     * featParam: List of types of features that will be generated. N is
%       the number of features of the type to be generated.
%        + 'LP':    Nx1 array of facts.
%        + 'HP':    Nx2 array of upper and lower facts.
%        + 'Lap':   Nx1 array of facts.
%        + 'Shear': number of bands on each level, including the coarse
%          level. 
%     * spBySize: Boolean. True if the partition of the labels and features
%       in data should be done by partitioning them into as many pieces
%       as possible, of the specified size. (Split By Size)
%     * spByQuan: Boolean. True if the labels and features in data should
%       be partitioned in a specified number of sets, as large as possible.
%       (Split By Quantity)
%     * prtSiz: Size of each part, if spBySize is true.
%     * numPrt: Number of parts, if spByQuan is true.
%     * numExp: Number of experiments to be ran.
% 
% OUTPUT:
%  - data: This struct contains the different elements necessary to run
%    an experiment.
%     * labels: Cell of Nx1 arrays containing the labels.
%     * feats: Cell of NxF arrays containing the respective features.
%     * bounds: Normalization bounds for each feature.
%     * F: Partitioned cell of feature matrices.
%     * L: Partitioned cell of label arrays.
%     * N: Number of experiment to run.
%     * s: Number or samples on each piece of features.
%     * n: Number of features pieces.
%     * C: Cost parameters for experiments.
%     * G: Gamma parameters for experiments.
%     * M: Index of the set of features to be used to generate the model
%          for each experiment.
% 

K = length(data.nameList);
theMins = cell(K,1);
theMaxs = cell(K,1);

for k = 1:K
    theSolid = RAWfromMHD(data.nameList{k},[],data.pathList{k});
    theSizes = size(theSolid);
    mB = modBox(data.boxes{k},theSizes);
    B = data.boxes{k};
    N = size(mB,1);
    pos = false(theSizes);
    neg = false(theSizes);
    pos(data.pTrainIdx{k}) = true;
    neg(data.nTrainIdx{k}) = true;
    tPos = false(size(pos));
    tNeg = false(size(neg));
    L = [];
    F = [];
    I = [];
    MM = cell(N,1);
    for n = 1:N
        tPos(:) = false;
        tNeg(:) = false;
        tPos(B(n,1):B(n,2),B(n,3):B(n,4),B(n,5):B(n,6)) = ...
            pos(B(n,1):B(n,2),B(n,3):B(n,4),B(n,5):B(n,6));
        tNeg(B(n,1):B(n,2),B(n,3):B(n,4),B(n,5):B(n,6)) = ...
            neg(B(n,1):B(n,2),B(n,3):B(n,4),B(n,5):B(n,6));
        sPos = tPos(mB(n,1):mB(n,2),mB(n,3):mB(n,4),mB(n,5):mB(n,6));
        sNeg = tNeg(mB(n,1):mB(n,2),mB(n,3):mB(n,4),mB(n,5):mB(n,6));
        SS = theSolid(mB(n,1):mB(n,2),mB(n,3):mB(n,4),mB(n,5):mB(n,6));
        P = find(sPos);
        N = find(sNeg);
        II = [P; N];
        LL = [P*0+1; N*0-1];
        [FF MM{n}] = genFeats(SS,II,data);
        P = find(tPos);
        N = find(tNeg);
        II = [P; N];
        L = cat(1,L,LL);
        F = cat(1,F,FF);
        I = cat(1,I,II);
    end
    theMins{k} = min(cat(3,MM{:}),[],3);
    theMaxs{k} = max(cat(3,MM{:}),[],3);
    data.labels{k} = L;
    data.feats{k} = F;
    data.indices{k} = I;
end

mini = min(cat(3,theMins{:}),[],3);
maxi = max(cat(3,theMaxs{:}),[],3);
M = zeros(size(mini));
M(1,:) = mini(1,:);
M(2,:) = maxi(2,:);

if ~isfield(data,'bounds')
    data.bounds = M;
end

data = prepareExperiment(data);

end


function exp = prepareExperiment(exp)
% This function takes the experiment struct and the experiment-options
% struct, and partitions the features and labels according to the options
% chosen by the user. It is assumed that features and labels have already
% been generated, and that they match appropriately.
% 
% INPUT:
%  * exp: Experiment struct containing labels, features and options.
% 
% OUPUT:
%  * exp: Experiment modified according to options.
% 

% Fixed parameters.
BIG = 100;  % The Gamma and Costs arrays are generated with values that are
RAT = 10;   % between 1/RAT and BIG/RAT, with 1/RAT spacing.
DEF = 2000; % Default size of pieces.
EXP = 1000; % Default number of experiments.

exp = countFeat(exp);

for n = 1:length(exp.feats);
    for k = 1:exp.totFeat
        M = exp.bounds(1,k);
        m = exp.bounds(2,k);
        V = exp.feats{n}(:,k);
        V = (V-m)/(M-m);
        exp.feats{n}(:,k) = V;
    end
end

I = cat(1,exp.indices{:});
L = cat(1,exp.labels{:});
F = cat(1,exp.feats{:});
M = length(L);
II = randperm(M).';


if exp.spBySize && exp.spByQuan
    exp.s = exp.prtSiz;
    if exp.numPrt < floor(M/exp.s)
        exp.n = exp.numPrt;
    else
        exp.n = floor(M/exp.s);
    end
elseif exp.spBySize
    exp.s = exp.prtSiz;
    exp.n = floor(M/exp.s);
elseif exp.spByQuan
    exp.n = exp.numPrt;
    exp.s = floor(M/exp.n);
else
    exp.s = DEF;
    exp.n = floor(M/exp.s);
end

if isfield(exp,'numExp')
    exp.N = exp.numExp;
else
    exp.N = EXP;
end

exp.F = cell(exp.n,1);
exp.L = cell(exp.n,1);
exp.G = randi(BIG,exp.N,1)/RAT;
exp.C = randi(BIG,exp.N,1)/RAT;
exp.M = randi(exp.n,exp.N,1);

for k = 1:exp.n
    exp.I{k} = I(II(((k-1)*exp.s+1):(k*exp.s)));
    exp.L{k} = L(II(((k-1)*exp.s+1):(k*exp.s)));
    exp.F{k} = F(II(((k-1)*exp.s+1):(k*exp.s)),:);
end

end


function opt = countFeat(opt)
% This is a small subroutine that counts the amount of features, both total
% and individual. It will also compute the shifted cumulative sum of the
% number of features.
% 
% INPUT:
%  - opt: Struct with a bunch of options to be counted and expanded.
% 
% OUTPUT: 
%  - opt: Struct of the expanded options.
%

n = length(opt.featParam);
opt.featN = zeros(n,1);

for k = 1:n
    switch opt.featType{k}
        case{'LP'}
            opt.featN(k) = length(opt.featParam{k});
        case{'Lap'}
            opt.featN(k) = length(opt.featParam{k});
        case{'HP'}
            opt.featN(k) = size(opt.featParam{k},1);
        case{'Shear'}
            opt.featN(k) = 3*sum(opt.featParam{k}.^2);
    end
end

opt.totFeat = sum(opt.featN);
opt.cumFeat = circshift(cumsum(opt.featN),1);
opt.cumFeat(1) = 0;

end


function mB = modBox(B,S)
% This function takes regions coordinates from a solid, and the size of the
% solid. It returns a set of modified set of regions, where a predefine
% padding is done.
% 
% INPUT:
%  - R: Cell containing a set of regions.
%  - S: Size of the solid.
% 
% OUTPUT:
%  - mR: Cell containing a set of modified regions.
% 

N = size(B,1);
mB = zeros(size(B));

for n = 1:N
    [xMin xMax] = adjBds(B(n,1),B(n,2),S(1));
    [yMin yMax] = adjBds(B(n,3),B(n,4),S(2));
    [zMin zMax] = adjBds(B(n,5),B(n,6),S(3));
    mB(n,:) = [xMin xMax yMin yMax zMin zMax];
end

end


function [m M] = adjBds(b,B,s)
% This function adjusts the bounds of a set of bounds according to a
% predifine padding rate.
% 
% INPUT:
%  - b, B: Original minimal and maximal bounds.
%  - s: maximal size of the respective dimension.
% 
% OUTPUT:
%  - m, M: minimal and maximal adjusted bounds.
% 

PAD = 0.1;

r = ceil(PAD*abs(B-b));
m = b - r;
M = B + r;
if m < 1
    m = 1;
end

if M > s
    M = s;
end

end

% CREATED: 
% - Date: 09/06/2012
% - By: David Jimenez.
% 
