% Copyright  2012 Computational Biomedicine Lab (CBL), 
% University of Houston. All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, is prohibited without the prior written consent of CBL.
%

function [feats,M] = genFeats(theSolid,theIdx,opt)
% function opt = genFeats(opt)
% This function receives a 3D solid, the set of indices for which the
% features have to be generated, and the set of options indicating what
% features should be generated. 
% 
% INPUT:
%  - theSolid: This is the three dimensional solid.
%  - theIdx: This is the set of indices for which the features will be
%            stored.
%  - opt: This is a struct containing two cell fields: 
%     * featType, that contains a list of the types of features: 'LP' (Low
%       Pass), 'HP' (High Pass), 'Lap' (Laplacian), 'Shear' (Shearlets). 
%     * featParam, For 'LP' or 'Lap', the respective parameter is an Nx1
%       array, where each element is the respective fact. For 'HP' it is an
%       Nx2  array, where the entry (i,1) is the upper fact for the i-th
%       feature, and the entry (i,2) is the lower fact. For 'Shear', it is
%       the number of bands on each level, including the coarse level.
%     * cubeSize: In the case that 
% 
% OUTPUT:
%  - feats: This is a function 
% 

opt =  countFeat(opt);
feats = zeros(length(theIdx),opt.totFeat,'single');
nFeatTypes = length(opt.featParam); % Number of feature types.
M = zeros(2,opt.totFeat,'single');

for k = 1:nFeatTypes
    n = opt.cumFeat(k) + 1;
    if k < nFeatTypes
        N = opt.cumFeat(k+1);
    else
        N = opt.totFeat;
    end
    switch opt.featType{k}
        case{'LP'}
            [TF, mM] = genLP(theSolid,theIdx,opt.featParam{k});
            M(:,n:N) = mM;
            feats(:,n:N) = TF;
        case{'Lap'}
            [TF, mM] = genLap(theSolid,theIdx,opt.featParam{k});
            M(:,n:N) = mM;
            feats(:,n:N) = TF;
        case{'HP'}
            [TF, mM] = genHP(theSolid,theIdx,opt.featParam{k});
            M(:,n:N) = mM;
            feats(:,n:N) = TF;
        case{'Shear'}
            [TF, mM]=genShear(theSolid,theIdx,opt.featParam{k},...
                opt.cubeSize);
            M(:,n:N) = mM;
            feats(:,n:N) = TF;
    end
end

end


function [tempFeat, mM] = genShear(S,I,B,C)
% This function generates the a set of Shearlet filtered features.
% 
% INPUT: 
%  - S: The 3D solid to be filtered.
%  - I: The set of indices for which the features will be stored.
%  - P: Bands per level.
%  - C: The cube size
% 
% OUTPUT:
%  - tempFeat: The matrix where the features will be generated.
%  - mM: A 2xN array containing maximum and minimum of the feature.
% 

L = length(B) - 1; % Level.
N = 3*sum(B.^2);
tempFeat = zeros(length(I),N,'single');
mM = zeros(2,N,'single');
% mM = zeros(N,2,'single');

F = meyerFilter3D(L+1,C,B);
P = radPyrDec(S,L);

currentFeature = 1;
for l = 1:(L+1)
    for k = 1:3
        [II JJ] = size(F{k,l});
        for i = 1:II
            for j = 1:JJ
                TF = convnfft(P{l},F{k,l}{i,j},'same');
                M = max(TF(:));
                m = min(TF(:));
                mM(1,currentFeature) = M;
                mM(2,currentFeature) = m;
                tempFeat(:,currentFeature) = TF(I);
                currentFeature = currentFeature + 1;
            end
        end
    end
end

end


function [tempFeat, mM] = genLap(S,I,P)
% This function generates the a set of Laplacian filtered features.
% 
% INPUT: 
%  - S: The 3D solid to be filtered.
%  - I: The set of indices for which the features will be stored.
%  - P: The set of parameters.
% 
% OUTPUT:
%  - tempFeat: The matrix where the features will be generated.
%  - mM: A 2xN array containing maximum and minimum of the feature.
% 

tempFeat = zeros(length(I), length(P), 'single');
mM = zeros(2,length(P), 'single');
[Nx Ny Nz] = size(S);
SF = fftn(S);

for i = 1:length(P)
    F = makeLapFilter(Nx,Ny,Nz,P(i));
    TF = SF.*F;
    TF = ifftn(TF);
    m = min(TF(:));
    M = max(TF(:));
    mM(1,i) = M;
    mM(2,i) = m;
    tempFeat(:,i) = TF(I);
end

end


function [tempFeat, mM] = genLP(S,I,P)
% This function generates the a set of Low Pass filtered features.
% 
% INPUT: 
%  - S: The 3D solid to be filtered.
%  - I: The set of indices for which the features will be stored.
%  - P: The set of parameters.
% 
% OUTPUT:
%  - tempFeat: The matrix where the features will be generated.
%  - mM: A 2xN array containing maximum and minimum of the feature.
% 

tempFeat = zeros(length(I), length(P), 'single');
mM = zeros(2,length(P), 'single');
[Nx Ny Nz] = size(S);
SF = fftn(S);

for i = 1:length(P)
    F = makeFilter(Nx,Ny,Nz,P(i));
    TF = SF.*F;
    TF = ifftn(TF);
    m = min(TF(:));
    M = max(TF(:));
    mM(1,i) = M;
    mM(2,i) = m;
    tempFeat(:,i) = TF(I);
end

end


function [tempFeat, mM] = genHP(S,I,P)
% This function generates the a set of Low Pass filtered features.
% 
% INPUT: 
%  - S: The 3D solid to be filtered.
%  - I: The set of indices for which the features will be stored.
%  - P: The set of parameters.
% 
% OUTPUT:
%  - tempFeat: The matrix where the features will be generated.
%  - mM: A 2xN array containing maximum and minimum of the feature.
% 

tempFeat = zeros(length(I), size(P,1), 'single');
mM = zeros(2, size(P,1), 'single');
[Nx Ny Nz] = size(S);
SF = fftn(S);

for i = 1:size(P,1)
    F = makeFilter(Nx,Ny,Nz,P(i,1)) - makeFilter(Nx,Ny,Nz,P(i,2));
    TF = SF.*F;
    TF = ifftn(TF);
    m = min(TF(:));
    M = max(TF(:));
    mM(1,i) = M;
    mM(2,i) = m;
    tempFeat(:,i) = TF(I);
end

end


function opt = countFeat(opt)
% This is a small subroutine that counts the amount of features, both total
% and individual. It will also compute the shifted cumulative sum of the
% number of features.
% 
% INPUT:
%  - opt: Struct with a bunch of options to be counted and expanded.
% 
% OUTPUT: 
%  - opt: Struct of the expanded options.
%

n = length(opt.featParam);
opt.featN = zeros(n,1);

for k = 1:n
    switch opt.featType{k}
        case{'LP'}
            opt.featN(k) = length(opt.featParam{k});
        case{'Lap'}
            opt.featN(k) = length(opt.featParam{k});
        case{'HP'}
            opt.featN(k) = size(opt.featParam{k},1);
        case{'Shear'}
            opt.featN(k) = 3*sum(opt.featParam{k}.^2);
    end
end

opt.totFeat = sum(opt.featN);
opt.cumFeat = circshift(cumsum(opt.featN),1);
opt.cumFeat(1) = 0;

end


function filt=makeLapFilter(nx,ny,nz,fact)
% This is the routine that generates filters for Laplacian filters.
% 
% INPUT:
%  - nx, ny, nz: Dimensions of the three dimensional filter.
%  - fact: Fact parameters for the filters.
% 
% OUTPUT:
%  - filt: The filter needed.
% 

% Determine maximum frequencies
kxmax = pi; kymax = pi; kzmax = pi;

% Determine Reflexion values
nhx=floor(nx/2); 
nhy=floor(ny/2); 
nhz=floor(nz/2);

flipx=mod(nx,2); 
flipy=mod(ny,2); 
flipz=mod(nz,2);

% Arrays
kx=(kxmax/nhx)*(0:nhx); 
ky=(kymax/nhy)*(0:nhy); 
kz=(kzmax/nhz)*(0:nhz); 

%HDAF parameters
ndaf=60;
kd=fact*kxmax;
sigma=sqrt(2.0*ndaf+1)/kd;

[Kx,Ky,Kz]=ndgrid(kx,ky,kz);
Kxy = Kx.^2+Ky.^2+Kz.^2;
filtro=Kxy.*hdaf(ndaf,sigma,Kxy);
filt = zeros(nx,ny,nz,'single');

% Filter reflexion.
filt(1:nhx+1,1:nhy+1,1:nhz+1) = filtro;
filt(nhx+2:nx,:,:) = filt(nhx+flipx:-1:2,:,:);
filt(:,nhy+2:ny,:) = filt(:,nhy+flipy:-1:2,:);
filt(:,:,nhz+2:nz) = filt(:,:,nhz+flipz:-1:2);

end


function filt = makeFilter(nx,ny,nz,fact)
% This is the routine that generates filters for Low Pass and High Pass
% filters.
% 
% INPUT:
%  - nx, ny, nz: Dimensions of the three dimensional filter.
%  - fact: Fact parameters for the filters.
% 
% OUTPUT:
%  - filt: The filter needed.
% 

% Determine maximum frequencies
kxmax = pi;
kymax = pi;
kzmax = pi;

% Determine Reflexion values
nhx=floor(nx/2);
nhy=floor(ny/2);
nhz=floor(nz/2);

flipx=mod(nx,2);
flipy=mod(ny,2);
flipz=mod(nz,2);

% Arrays
kx=(kxmax/nhx)*(0:nhx); 
ky=(kymax/nhy)*(0:nhy); 
kz=(kzmax/nhz)*(0:nhz); 

%HDAF parameters
ndaf = 60;
kd = fact * kxmax;
sigma = sqrt(2.0*ndaf+1)/kd;

% Filter reflexion.
[Kx,Ky,Kz] = ndgrid(kx,ky,kz);
Kxy = Kx.^2+Ky.^2+Kz.^2;
filtro = hdaf(ndaf,sigma,Kxy);
filt = single(zeros(nx,ny,nz));

% Filter reflexion.
filt(1:nhx+1,1:nhy+1,1:nhz+1) = filtro;
filt(nhx+2:nx,:,:) = filt(nhx+flipx:-1:2,:,:);
filt(:,nhy+2:ny,:) = filt(:,nhy+flipy:-1:2,:);
filt(:,:,nhz+2:nz) = filt(:,:,nhz+flipz:-1:2);

end


function F = hdaf(n,s,x)
% INPUT
%  - n: Number of Approximations for the Taylor Polynomial
%  - s: Deviation of the Gaussian (sigma)
%  - x: Values to evaluate. It may be a scalar, or an array.
% 
% OUTPUT
% \left[\sum_{i=0}^{n}\frac{1}{k!}\left(\frac{x\sigma^2}{2}\right)^{i}
% \right]e^{-\frac{x\sigma^2}{2}}
% 

  en = 1.0; 
  ft = 1.0;
  for i=1:n
    ft = ft*i;
    en = en+((x*s^2)/2).^i/ft;
  end
  F = single(en.*exp((-1.0)*x*(s^2)/2));
  
end


function P = radPyrDec(S,L)
% Radial Pyramidal Decomposition of the number of levels indicated
% 
% INPUT:
%  - S: The Solid in Question
%  - L: the number of levels of the decomposition.
% 
% OUTPUT:
%  - P: an L+1 by 1 cell array storyng the subbands from the finest to the
%       coarsest scale.
% 

S = single(S);
X = fftn(S);
[nx ny nz] = size(X);
uf = 1.3; % Upper adjustment factor. Works for ndaf = 60.
lf = 0.8; % Lower adjustment factor. Works for ndaf = 60.
P = cell(L+1,1);
for k = 1:(L+1)
    if k == 1
        fact1 = lf*L/(L+1);
        F = makeFilter(nx,ny,nz,fact1);
        F(isinf(F)) = 0;
        F(isnan(F)) = 0;
        F = 1 - F;
    elseif k == L+1
        fact2 = uf/(L+1);
        F = makeFilter(nx,ny,nz,fact2);
        F(isinf(F)) = 0;
        F(isnan(F)) = 0;
    else
        fact1 = uf*(L+2-k)/(L+1);
        fact2 = lf*(L+1-k)/(L+1);
        F1 = makeFilter(nx,ny,nz,fact1);
        F2 = makeFilter(nx,ny,nz,fact2);
        F1(isinf(F1)) = 0;
        F1(isnan(F1)) = 0;
        F2(isinf(F2)) = 0;
        F2(isnan(F2)) = 0;
        F = F1 - F2;
    end
    P{k} = single(ifftn(X.*F));
end

end


function F = meyerFilter3D(level,cubeSize,levBands)
% It generates a set of Meyer Based Filters in 3D, according to Pooran's
% code.
% 
% INPUT: 
%  - level: Number of level decomposition. Integer.
%  - cubeSize: The length of the filter for convolution. Integer
%  - levBands: the number of bands at each level. Integer array level-long.
% 
% OUTPUT:
%  - F: 3xlevel cell of cells, each subsell contains a set of band filters
%       for the indicated dimension (direction) and level.
%

F = cell(3,level);
for l = 1:level
    numDir = levBands(l);
    P = genPyrSection(cubeSize);%,'single');
    S = cubeSize/numDir; % variable shift from Pooran's
    A = zeros(cubeSize,cubeSize,cubeSize,'single');
    for c = 1:3
        for i = 1:numDir
            for j = 1:numDir
                % Define I, the size of R, and fill the 1s and 1/2s.
                if (i == numDir) && (j == numDir)
                    R = ones(S,cubeSize,S)/2;
                    R(2:(S-1),1:cubeSize,2:(S-1)) = 1;
                    I = [(j-1)*S+1, j*S; 1, cubeSize;(i-1)*S+1, i*S];
                elseif (i == numDir)
                    R = ones(S+1,cubeSize,S)/2;
                    R(2:S,1:cubeSize,2:(S-1)) = 1;
                    I = [(j-1)*S+1, j*S+1; 1, cubeSize;(i-1)*S+1, i*S];
                elseif (j == numDir)
                    R = ones(S,cubeSize,S+1)/2;
                    R(2:(S-1),1:cubeSize,2:S) = 1;
                    I = [(j-1)*S+1, j*S; 1, cubeSize;(i-1)*S+1, i*S+1];
                else
                    R = ones(S+1,cubeSize,S+1)/2;
                    R(2:S,1:cubeSize,2:S) = 1;
                    I = [(j-1)*S+1, j*S+1; 1, cubeSize;(i-1)*S+1, i*S+1];
                end
                
                % Fill up the 1/3s
                if (i == 1)&&(j == 1)
                    R(1, 1:cubeSize, 1) = 1/3;
                elseif (i == 1)&&(j == numDir)
                    R(S, 1:cubeSize, 1) = 1/3;
                elseif (i == numDir)&&(j == 1)
                    R(1, 1:cubeSize, S) = 1/3;
                elseif (i == numDir)&&(j == numDir)
                    R(S, 1:cubeSize, S) = 1/3;
                end
                F{c,l}{i,j} = polToRec(I,R,P{c});
                A = A + F{c,l}{i,j};
            end
        end
    end
    
    for c = 1:3
        for i = 1:numDir
            for j = 1:numDir
                F{c,l}{i,j} = F{c,l}{i,j}./A;
                F{c,l}{i,j} = single(real(fftshift(ifftn(fftshift(F{c,l}{i,j})))));
            end
        end
    end
    
end

end


function B = polToRec(radIdx,mRad,P)
% The transformation of the filter from polar to rectangular coordinates. I
% need a better understanding of how this works, and potentially, a way to
% make it more efficient. For now, I just modified Pooran's code to make it
% less redundant.
% 
% INPUT:
%  - radIdx: Coordinates of inference.
%  - mRad: Value.
%  - P: Filter in Polar Coordinates.
% 
% OUTPUT:
%  - B: Filter in rectangular coordinates.
%

B = single(zeros(size(P.X)));
for i = radIdx(1,1):radIdx(1,2)
    for j = radIdx(2,1):radIdx(2,2)
        for k = radIdx(3,1):radIdx(3,2)
            B(P.X(i,j,k),P.Y(i,j,k),P.Z(i,j,k)) = ...
                mRad(i-radIdx(1,1)+1,j-radIdx(2,1)+1,k-radIdx(3,1)+1);
        end
    end
end

end


function P = genPyrSection(n)
% This function generates the three dimentional cones. 
% 
% INPUT:
%  - n: cube size.
% 
% OUTPUT:
%  - P: Cell with the pyramidal sections.
% 

P = cell(1,3);
[P{1}.X P{1}.Y P{1}.Z]= genXYZ(n);
P{2}.X = P{1}.Y; P{2}.Y = P{1}.X; P{2}.Z = P{1}.Z;
P{3}.X = P{1}.Z; P{3}.Y = P{1}.Y; P{3}.Z = P{1}.X;

end


function [X,Y,Z] = genXYZ(n)
% This function generates sets of indices, according to Pooran Negi's
% algorithm.
% 
% INPUT:
%  - n: Cube size
% 
% OUTPUT:
%  - X: Cone in the X direction.
%  - Y: Cone in the Y direction.
%  - Z: Cone in the Z direction.
% 

[pI, pJ, pK] = ndgrid(1:n,1:n,1:n);
[nI, ~, nK] = ndgrid(n:-1:1,n:-1:1,n:-1:1);
X = uint16(pJ);
Y = uint16(pI + (pJ-1).*abs(pI-nI)/(n-1));
Z = uint16(pK + (pJ-1).*abs(pK-nK)/(n-1));
mP=ceil(n/2);
pDir = 1:mP;
nDir = n+1-pDir;
Y(nDir,n:-1:1,1:n) = Y(pDir,1:n,1:n);
Z(1:n,n:-1:1,nDir) = Z(1:n,1:n,pDir);

end


% CREATED: 
% - Date: 09/06/2012
% - By: David Jimenez.
% 
