% Copyright  2012 Computational Biomedicine Lab (CBL), 
% University of Houston. All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, is prohibited without the prior written consent of CBL.
%

function data = genIdx(data)
% This function generates a set of indices according to the options given
% by data struct (Indices Options). It may generate training indices,
% prediction indices, or both. 
% 
% INPUT:
%   - data: Struct. It contains different options, depending on what the
%     process needed. The following are the fields of the struct
%      * trainIdx: Boolean. True if training indices should be generated.
%      * predIdx: Boolean. True if prediction indices should be generated.
%      * genLap: Boolean. True if there is the need of generating the
%        Laplacian solids needed for the training indices. If false, it is
%        assumed that such solids have already been generated. 
%      * posLap: 3D Array. If useList is false and genLap is true, then 
%        this is the filtered copy used for positive coefficients.
%      * negLap: 3D Array. If useList is false and genLap is true, then 
%        this is the filtered copy used for negative coefficients.
%      * pathList: Cell. It useList is true, this field contains the list
%        of paths for each solid. One path per solid, even if all solids
%        are contained in the same path.
%      * nameList: Cell. It useList is true, this field contains the list
%        of name for each solid.
%      * saveLap: Boolean. If nameList and pathList is define, and this
%        field is true, it saves the produced Laplacian solids.
%      * lapFact: Array, Kx1. Contains the facts to be used for the
%        algorithm to generate the negative and positive Laplacian solids.
%      * boxes: Cell of Kx6 Arrays. Contains the coordinates of the boxes 
%        delimiting the sections of the solid from where the samples will
%        be taken.
%      * pos: Cell of Kx2 Arrays. Contains posP and negP sets of parameters
%        for indices generation.
%      * saveTrain: Boolean. True if positive and negative samples should
%        be saved to solids on the hard drive.
%      * predStr: String. Tells what type of method to use to generate the
%        predicting indices
%         + 'MEAN': All voxels above the average in raw solid.
%         + 'PERC': All voxels with values in the upper specified
%                   percentage.
%         + 'ENSF': Envelope around surface around provided binary.
%      * predPerc: Percentage of the solid to obtain predicting indices.
%      * binary: Binary solid to obtain second run of segmentation.
%      * eps: Epsilon necessary for envelops.
%      * binNameList: Cell. Used for the names of the binary solids.
% 
% OUTPUT:
%  - data: Struct. It contains the indices required
%      * pTrainIdx: Cell. Contains arrays of positive training indices.
%      * nTrainIdx: Cell. Contains arrays of negative training indices.
%      * predictIdx: Predicting Indices.
%      * regions: Cell of regions for the second run. Each region is
%                 represented by a 3x2 cell, with the coordinates of
%                 maximum and minimum vertices of it.
% 

% If we are using a list of solids.
K = length(data.nameList);
data.pTrainIdx = cell(K,1);
data.nTrainIdx = cell(K,1);
data.predictIdx =  cell(K,1);

for k = 1:K
    
    theSolid = RAWfromMHD(data.nameList{k},[],data.pathList{k});
    
    if data.trainIdx && data.genLap
        % If there is the need of
        theSizes = size(theSolid);
        [posLap negLap] = genLapPosNeg(theSolid,data.lapFact);
        if data.saveLap % Laplacian solids to be saved.
            % If it is needed to save the files.
            saveLapIdxSolids(posLap, negLap, data.pathList{k},...
                data.nameList{k});
        end
        [P N] = idxLap(theSolid, posLap,negLap,data.boxes{k},...
            data.pos{k});
        data.pTrainIdx{k} = P;
        data.nTrainIdx{k} = N;
        if data.saveTrain % Training samples to be saved.
            saveIndexSolid(P, N, data.pathList{k}, ...
                data.nameList{k}, theSizes);
        end
    elseif data.trainIdx
        thePath = data.pathList{k};
        pName = strcat(data.nameList{k},'_PLAP');
        nName = strcat(data.nameList{k},'_NLAP');
        posLap = RAWfromMHD(pName,[],thePath);
        negLap = RAWfromMHD(nName,[],thePath);
        theSizes = size(theSolid);
        [P N] = idxLap(theSolid, posLap,negLap,data.boxes{k},...
            data.pos{k});
        data.pTrainIdx{k} = P;
        data.nTrainIdx{k} = N;
        if data.saveTrain % Training samples to be saved.
            saveIndexSolid(P, N, data.pathList{k}, ...
                data.nameList{k}, theSizes);
        end
    end
    
    % If there is a need to generate
    if data.predIdx
        switch data.predStr
            case{'MEAN'}
                theSolid = RAWfromMHD(data.nameList{k},[],...
                    data.pathList{k});
                data.predictIdx{k} = overMeanIdx(theSolid);
            case{'PERC'}
                theSolid = RAWfromMHD(data.nameList{k},[],...
                    data.pathList{k});
                data.predictIdx{k} = percentIdx(theSolid,data.predPerc);
            case{'ENSF'}
                theBinary = RAWfromMHD(data.binNameList{k},[],...
                    data.pathList{k});
                data.predictIdx{k} = surfaceEnvelope(theBinary,data.eps);
                data.regions{k} = regFromBin(theBinary,data.predictIdx{k});
        end
    end
    
end

end


function R = regFromBin(B,I)
% This function returns the set of regions needed to process for the second
% pass of the ORION II processing.
%
% INPUT:
%  - B: Binary solid obtained from the first pass.
%  - I: Set of indices that will be considered in second run.
%
% OUTPUT:
%  - R: Set of regions. Each region is returned simply as the coordinates
%       of the two maximal corners of the entire subvolume.
%

TOT = 22;  % The log_2 of the maximum allowable volume
FRA = 7/8; % This is the fraction of the side of the subsolids that
% separates the centers of different

B(I) = true;
h = size(B,3);
lh = ceil(log2(h));
r = 2^(floor((TOT-lh)/2)-1);
d = round(FRA*r);


B = max(B,[],3);

[Nx Ny] = size(B);
X = d:2*d:Nx;
Y = d:2*d:Ny;

n = 1;
R{n} = [];

for i = 1:length(X)
    for j = 1:1:length(Y)
        [xMin xMax] = arrVertex(X(i),r,Nx);
        [yMin yMax] = arrVertex(Y(j),r,Ny);
        J = B(xMin:xMax,yMin:yMax);
        if ~isempty(find(J,1))
            R{n} = [xMin xMax;yMin yMax; 1 h];
            n = n + 1;
        end
        
    end
end

end


function I = surfaceEnvelope(B,r)
% This function returns the indices from a binary solid. It basically takes
% every voxel within r voxels from the surface.
%
% INPUT:
%  - B: Binary Solid
%  - r: radius.
%
% OUTPUT:
%  - I: Set of indices
%

B = logical(B);
J = find(B);
N = length(J);
S = size(B); % Sizes
A = false(S);
for n = 1:N
    [i,j,k] = ind2sub(S,J(n));
    [mx Mx] = arrVertex(i,r,S(1));
    [my My] = arrVertex(j,r,S(2));
    [mz Mz] = arrVertex(k,r,S(3));
    T = B(mx:Mx,my:My,mz:Mz);
    if numel(T) > numel(T(T))
        A(mx:Mx,my:My,mz:Mz) = true;
    end
end
I = find(A);

end


function I = percentIdx(S,P)
% This function returns the indices of the voxels with values in the upper
% provided percent of in the given solid raw solid.
%
% INPUT:
%  - S: The Raw Solid.
%  - P: Percentage to be taken.
%
% OUTPUT:
%  - I: Set of Indices.
%

[V D] = sort(S(:));
T = length(V);
t = round(T*P/100);
I = D(T-t:T);

end


function I = overMeanIdx(S)
% This function returns the indices of the voxels in the given solid that
% are above the average.
%
% INPUT:
%  - S: The Raw Solid.
%
% OUTPUT:
%  - I: Set of Indices.
%

M = mean(mean(mean(S)));
I = find(S > M);

end


function saveIndexSolid(P,N,path,name,sizes)
% This function saves the index solids in the drive.
% 
% INPUT:
%  - P: Array of positive indices.
%  - N: Array of negative indices. 
%  - path: Path in storage where to save the solids.
%  - name: The name of the file to save in storage.
%  - sizes: List of the dimensions of the solid.
% 

pName = strcat(name,'_P');
nName = strcat(name,'_N');
PP = false(sizes);
NN = false(sizes);
PP(P) = true;
NN(N) = true;
WriteRAWandMHD(PP,pName,path);
WriteRAWandMHD(NN,nName,path);

end


function saveLapIdxSolids(P,N,path,name)
% Plain and simple. This function saves the Laplacian filtered solids
% needed to generate the training solids.
% 
% INPUT:
%  - P: Positive Laplacian solid.
%  - N: Negative Laplacian solid.
%  - path: Path in storage where to save the solids.
%  - name: The name of the file to save in storage.
% 

pName = strcat(name,'_PLAP');
nName = strcat(name,'_NLAP');
WriteRAWandMHD(P,pName,path);
WriteRAWandMHD(N,nName,path);

end


function [P N] = idxLap(S,pL,nL,boxes,pos)
% This function generates the positive and negative samples from the solid
% and the Laplacian filtered copies.
%
% INPUT:
%  - S: Raw Solid.
%  - pL: Positive Laplacian filtered solid.
%  - nL: Negative Laplacian filtered solid.
%  - boxes: Kx6 array, containing the coordinates of the boxes to consider.
%  - pos: Kx2 array, containing the negP and posP parameters for each box.
%
% OUTPUT:
%  - P: Positive Indices.
%  - N: Negative Indices.
%

PS = 0.1; % part of the solid that would be taken from the original

bigN = false(size(S));
bigP = false(size(S));

K = size(pos,1);
M = mean(mean(mean(S)));

for k = 1:K
    posP = pos(k,1) / 100;
    negP = pos(k,2) / 100;
    xMin = boxes(k,1);
    xMax = boxes(k,2);
    yMin = boxes(k,3);
    yMax = boxes(k,4);
    zMin = boxes(k,5);
    zMax = boxes(k,6);
    
    tempSol = pL(xMin:xMax,yMin:yMax,zMin:zMax);
    
    tempP = false(size(tempSol));
    tempN = false(size(tempSol));
    tempSize = numel(tempSol);
    tempPosP = round(tempSize*posP);
    tempPosN = round(tempSize*negP);
    
    tempSol = nL(xMin:xMax,yMin:yMax,zMin:zMax);
    [V I] = sort(tempSol(:));
    T = length(V);
    tempP(I(T-tempPosP:T)) = true;
    
    tempSol = pL(xMin:xMax,yMin:yMax,zMin:zMax);
    [V I] = sort(tempSol(:));
    tempN(I(1:tempPosN)) = true;
    length(V);
    
    tempSol = S(xMin:xMax,yMin:yMax,zMin:zMax);
    [V I] = sort(tempSol(:));
    tempP(I(end-round(tempPosP*PS):end)) = true;
    F = find(V>M,1);
    tempN(I(F-round(tempPosN*PS/2):F+round(tempPosN*PS/2))) = true;
    
    bigN(xMin:xMax,yMin:yMax,zMin:zMax) = tempN;
    bigP(xMin:xMax,yMin:yMax,zMin:zMax) = tempP;
end

P = find(bigP);
N = find(bigN);

end


function [P N] = genLapPosNeg(Sol,fct)
% This function generates the filtered versions for the negative and
% positive indices.
%
% INPUT:
%  - Sol: Original raw solid.
%  - fct: List of facts. Kx1. Contains the list of facts.
%
% OUTPUT:
%  - P: Positive Laplacian solid.
%  - N: Negative Laplacian solid.
%

[Nx Ny Nz] = size(Sol);
K = length(fct);
S = fftn(Sol);
P = zeros(Nx,Ny,Nz,'single');
N = zeros(Nx,Ny,Nz,'single');
TN = zeros(Nx,Ny,Nz,2,'single');
TP = zeros(Nx,Ny,Nz,2,'single');

for k = 1:K
    F = makeLapFilter(Nx,Ny,Nz,fct(k));
    T = S.*F;
    T = ifftn(T);
    if k == 1
        N = T;
        P = T;
    else
        TN(:,:,:,2) = T;
        TP(:,:,:,2) = T;
        N = min(TN,[],4);
        P = max(TP,[],4);
    end
    TN(:,:,:,1) = N;
    TP(:,:,:,1) = P;
end

end


function filt = makeLapFilter(nx,ny,nz,fact)
% This is the routine that generates filters for Laplacian filters.
%
% INPUT:
%  - nx, ny, nz: Dimensions of the three dimensional filter.
%  - fact: Fact parameters for the filters.
%
% OUTPUT:
%  - filt: The filter needed.
%

% Determine maximum frequencies
kxmax = pi; kymax = pi; kzmax = pi;

% Determine Reflexion values
nhx=floor(nx/2);
nhy=floor(ny/2);
nhz=floor(nz/2);

flipx=mod(nx,2);
flipy=mod(ny,2);
flipz=mod(nz,2);

% Arrays
kx=(kxmax/nhx)*(0:nhx);
ky=(kymax/nhy)*(0:nhy);
kz=(kzmax/nhz)*(0:nhz);

%HDAF parameters
ndaf=60;
kd=fact*kxmax;
sigma=sqrt(2.0*ndaf+1)/kd;

[Kx,Ky,Kz]=ndgrid(kx,ky,kz);
Kxy = Kx.^2+Ky.^2+Kz.^2;
filtro=Kxy.*hdaf(ndaf,sigma,Kxy);
filt = zeros(nx,ny,nz,'single');

% Filter reflexion.
filt(1:nhx+1,1:nhy+1,1:nhz+1) = filtro;
filt(nhx+2:nx,:,:) = filt(nhx+flipx:-1:2,:,:);
filt(:,nhy+2:ny,:) = filt(:,nhy+flipy:-1:2,:);
filt(:,:,nhz+2:nz) = filt(:,:,nhz+flipz:-1:2);

end


function F = hdaf(n,s,x)
% INPUT
%  - n: Number of Approximations for the Taylor Polynomial
%  - s: Deviation of the Gaussian (sigma)
%  - x: Values to evaluate. It may be a scalar, or an array.
% 
% OUTPUT
% \left[\sum_{i=0}^{n}\frac{1}{k!}\left(\frac{x\sigma^2}{2}\right)^{i}
% \right]e^{-\frac{x\sigma^2}{2}}

  en = 1.0; 
  ft = 1.0;
  for i=1:n
    ft = ft*i;
    en = en+((x*s^2)/2).^i/ft;
  end
  F = single(en.*exp((-1.0)*x*(s^2)/2));
  
end


function [m M] = arrVertex(x,rr,N)
% This simple function verifies that the box in question has all the
% vertices inside the solid.
% 
% INPUT
%  - x: Center coordinate of the box.
%  - rr: Radius of the box in question.
%  - N: Maximum possible coordinate
% 
% OUTPUT:
%  - m: Actual minimal coordinate of the box.
%  - M: Actual maximal coordinate of the box.
% 

if x-rr < 1
    m = 1;
else
    m = x-rr;
end

if x+rr > N
    M = N;
else
    M = x+rr;
end

end

% CREATED: 
% - Date: 09/06/2012
% - By: David Jimenez.
% 
