% Copyright  2012 Computational Biomedicine Lab (CBL), 
% University of Houston. All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, is prohibited without the prior written consent of CBL.
%

function exp = runExperiment(exp)
% This is a function to run a generated experiment (see genExperiment.m).
% 
% INPUT:
%  - exp: Experiment structure, as generated by genExperiment. 
%     * F: Partitioned cell of feature matrices.
%     * L: Partitioned cell of label arrays.
%     * N: Number of experiment to run.
%     * s: Number or samples on each piece of features.
%     * n: Number of features pieces.
%     * C: Cost parameters for experiments.
%     * G: Gamma parameters for experiments.
%     * M: Index of the set of features to be used to generate the model
%          for each experiment.
% 
% OUTPUT:
%  - exp: The experiment structure, when the experiments have already been
%    ran. The added fields are:
%     * SupVC: Array with the number of support vectors per experiment.
%     * ModAc: Array with the model accuracy per experiment.
%     * falPs: Array with the number of false positives per experiment
%     * falNg: Array with the number of false 
%     * bestG: Array with the best SAV# of gammas.
%     * bestC: Array with the best SAV# of costs.
% 

% Fixed parameters.
SAV = 3; % Number of optimal parameters to save

mL = zeros(exp.N,1);
mA = zeros(exp.N,exp.n);
fP = zeros(exp.N,exp.n);
fN = zeros(exp.N,exp.n);
C = exp.C;
G = exp.G;
MN = exp.M;

parfor k = 1:exp.N
    optSt = ['-b 1 -c ' num2str(C(k)) ' -g ' num2str(G(k))];
    M = svmtrain(double(exp.L{MN(k)}), double(exp.F{MN(k)}),optSt);
    mL(k) = length(M.sv_coef);
    tempA = zeros(1,exp.n);
    tempP = zeros(1,exp.n);
    tempN = zeros(1,exp.n);
    for n = 1:exp.n
        [pL,a,~] = svmpredict(double(exp.L{n}),double(exp.F{n}),M,'-b 1');
        tempA(n) = a(1);
        [fp fn] = falsePosNeg(double(exp.L{n}),pL);
        tempP(n) = fp;
        tempN(n) = fn;
    end
    mA(k,:) = tempA;
    fP(k,:) = tempP;
    fN(k,:) = tempN;
end

exp.SupVC = mL;
exp.ModAc = mA;
exp.falPs = fP;
exp.falNg = fN;

MA = mean(mA,2);
[~, J] = sort(MA,'descend');
I = J(1:SAV);

exp.bestG = exp.G(I);
exp.bestC = exp.C(I);
exp.bestM = exp.M(I);

end


function [fP fN] = falsePosNeg(tV,vV)
% Function for finding number of false positives and false negatives.
% 
% INPUT:
%  - tV: True Vector.
%  - vV: Validation Vector.
% 
% OUTPUT:
%  - fP: False Positive.
%  - fN: False Negative.
% 

fP = length(find((tV ~= vV)&(vV > 0)));
fN = length(find((tV ~= vV)&(vV < 0)));

end


% CREATED: 
% - Date: 09/06/2012
% - By: David Jimenez.
% 
