% Copyright  2012 Computational Biomedicine Lab (CBL), 
% University of Houston. All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, is prohibited without the prior written consent of CBL.
%

function runSegmentation(data)
% This function uses the information contained in the struct 'data' to find
% a solid and a model, and then generate from there the segmentation of the
% raw data.
% 
% INPUT:
%
% - data: Struct that contains the name and path of the model, as well as a
%         list of paths and names for the solids to segment.
%    * modPath: String. Path where the model is stored.
%    * modName: String. Name of the file of the model.
%    * pathList: Cell of strings. Contains the paths of the models.
%    * nameList: Cell of strings. Contains the file names of the models.
%      

% List of configurable constants.
PRT1 = 2^18;

% Load model, compute number of files to process, and options.
modLoad = fullfile(data.modPath,data.modName);
m = load(modLoad);
model = m.model;
K = length(data.nameList);
opt.featType = model.featType;
opt.featParam = model.featParam;

for k = 1:K
    % Loading solid 
    S = RAWfromMHD(data.nameList{k}, [], data.pathList{k});
    m = mean(mean(mean(S)));
    I1 = find(S > m);
    
    % Generate and normalize features
    F1 = genFeats(S,I1,opt);
    N = size(model.bounds,2);
    
    for n = 1:N
        M = model.bounds(1,n);
        m = model.bounds(2,n);
        V = F1(:,n);
        V = (V-m)/(M-m);
        F1(:,n) = V;
    end
    
    N = floor(numel(I1)/PRT1);
    I = cell(N+1,1);
    F = cell(N+1,1);
    P = cell(N+1,1);
    L = cell(N+1,1);
    
    % Slice variable
    for n = 0:N
        LB = n*PRT1+1;
        if n == N
            UB = numel(I1);
        else
            UB = (n+1)*PRT1;
        end
        I{n+1} = double(I1(LB:UB));
        F{n+1} = double(F1(LB:UB,:));
    end
    
    clear F1;
    
    % Running prediction for first model.
    parfor n = 1:(N+1)
        [LL, ~, PP] = svmpredict(I{n}, F{n}, model.svm, '-b 1');
        P{n} = PP;
        L{n} = LL;
    end
    
    I1 = cat(1,I{:});
    P1 = cat(1,P{:});
    L1 = cat(1,L{:});
    
    clear I P F L;
    
    B = false(size(S));
    B(I1) = (L1+1)/2;
    S = zeros(size(S),'single');
    
    if model.svm.Label(1) == 1
        S(I1) =  P1(:,1);
    else
        S(I1) =  P1(:,2);
    end
    
    saveName = [data.nameList{k} model.suffix '_BIN'];
    prbName = [data.nameList{k} model.suffix '_PRB'];
    WriteRAWandMHD(S,prbName,data.pathList{k});
    WriteRAWandMHD(B,saveName,data.pathList{k});
	
end


end

% CREATED: 
% - Date: 09/06/2012
% - By: David Jimenez.
% 
