% Copyright  2012 Computational Biomedicine Lab (CBL), 
% University of Houston. All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, is prohibited without the prior written consent of CBL.
%

function [D I] = viablePairs(S)
% This is an algorithm that takes the information of the connected
% components struct, and returns two arrays with the indices of pairs that
% seem to be viable, according to constants RUN1 and RUN2.
% 
% INPUT:
%  - S: Struct as returnd by connComp.
% 
% OUTPUT:
%  - D: Array of distances.
%  - I: Indices of pair of distances.
% 
% NOTES:
%  - The value of I(K) indicates that if [i j] = ind2sub(S.compNum, I(k)),
%    then the pair of connected components (i,j) is considered "viable",
%    this means we consider they may be connected.
%  - If [i j] = ind2sub(S.compNum, I(k)), then D(k) is the distance (or our
%    approximation) between the components i and j.
%  - IMPORTANT: We may want to not define RUN1 and RUN2 as hard constants,
%    but as constants included in S.
% 

% Constants:
% RUN1 = sqrt(27);
% RUN2 = sqrt(27);
RUN1 = 6;
RUN2 = sqrt(27);

R = cell(S.solDim,1);
T = find(logical(triu(ones(S.compNum),1)));

parfor k = 1:S.solDim
    I = single(S.coordMin(k,:));
    J = single(S.coordMax(k,:));
    [xMin yMin] = ndgrid(I,I);
    [xMax yMax] = ndgrid(J,J);
    I1 = (xMax < yMin);
    I2 = (yMax < xMin);
    C1 = yMin - xMax;
    C2 = xMin - yMax;
    M = zeros(S.compNum);
    M(I1) = C1(I1);
    M(I2) = C2(I2);
    R{k} = M;
end

M = max(cat(3,R{:}),[],3);
M = M(T);
IDX = T(M < RUN1);
E = NaN(size(IDX));
clear R T M xMin yMin xMax yMax I J;
K = length(IDX);

for k = 1:K
    [i j] = ind2sub(S.compNum,IDX(k));
    E(k) = distSet(S.compCoord{j}, S.compCoord{i});
end

J = find(E < RUN2);
I = IDX(J);
D = E(J);

end


function d = distSet(II,JJ)
% This function returns the distance, or an approximation of the distance,
% between the vectors contained in the vectors II and JJ. 
% 
% INPUT:
%  - II: dxM array of M d-dimensional vectors. 
%  - JJ: dxN array of N d-dimensional vectors. 
% 
% OUTPUT: 
%  - d: Distance or approximate between the sets of vectors in II and JJ.
% 

MM = 2^10; 
RT = 2^4;

Ni = size(II,2); 
Nj = size(JJ,2); 
D = [];
if (Ni > MM) && (Nj > MM)
    if (ceil((Ni*Nj)/(MM^2*RT)) > ceil(MM/RT))
        M = ceil((Ni*Nj)/(MM^2*RT));
    else
        M = min([ceil((Ni*Nj)/MM^2), ceil(MM/RT)]);
    end
    for k = 1:M
        I = randsample(Ni,MM);
        J = randsample(Nj,MM);
        d = dS(II(:,I),JJ(:,J));
        D = cat(1,D,d);
    end
elseif (Ni > MM)
    if (ceil(Ni/(MM*RT)) > ceil(MM/RT))
        M = ceil(Ni/(MM*RT));
    else
        M = min([ceil(Ni/MM), ceil(MM/RT)]);
    end
    for k = 1:M
        I = randsample(Ni,MM);
        d = dS(II(:,I),JJ);
        D = cat(1,D,d);
    end
elseif (Nj > MM)
    if (ceil(Ni/(MM*RT)) > ceil(MM/RT))
        M = ceil(Ni/(MM*RT));
    else
        M = min([ceil(Ni/MM), ceil(MM/RT)]);
    end
    for k = 1:M
        J = randsample(Nj,MM);
        d = dS(II,JJ(:,J));
        D = cat(1,D,d);
    end
else
    D = dS(II,JJ);
end

d = min(D);

end


function d = dS(A,B)
% Basic minimum distance function between the vectors contained in the
% matrix A and the vector contained in the vector B.
%
% INPUT:
%  - A: dxM array of M d-dimensional vectors. 
%  - B: dxN array of N d-dimensional vectors. 
% 
% OUTPUT: 
%  - d: Distance or approximate between the sets of vectors in A and B.
% 

AA = repmat(sum(A.*A).',[1 size(B,2)]);
BB = repmat(sum(B.*B),[size(A,2) 1]);
d = min(min(sqrt(abs(AA+BB-2*A'*B))));

end


% CREATED: 
% - Date: 09/06/2012
% - By: David Jimenez.
% 
